.IFNDEF WS28XX
.EQU WS28XX=1

;include generic stuff -------------------------------------------------------------------

.INCLUDE "orghelpers.inc"
.INCLUDE "unifiedaccess.inc"
.INCLUDE "unifiedmul.inc"

;include configuration -------------------------------------------------------------------

.EQU WS28XX_RGBFMT_RGB=0
.EQU WS28XX_RGBFMT_GRB=1

.INCLUDE "ws28xx_cfg.inc"

;helpful informations to understand what actually happens --------------------------------

;One color triple has three bytes, each of them eight bits.
.EQU WS28XX_CT_BITS = 3 * 8

;We use three physical SPI bits to encode one payload bit. The encoding
;patterns are 0b100 for a zero and 0b110 for one payload bit.
.EQU WS28XX_ENC_BITS = 3

;We need this amount of physical SPI bits for one color triple.
.EQU WS28XX_ENC_CT_BITS = WS28XX_CT_BITS * WS28XX_ENC_BITS

;In the encoded form one color triple occupies this number of bytes.
.EQU WS28XX_ENC_CT_BYTES = WS28XX_ENC_CT_BITS / 8

;for the reset/latch signal generation we emit a constant low for the time of
;n color triples. The "n" is declared here.
.EQU WS28XX_RST_NCT = 2

;informational: the clock tick counts ----------------------------------------------------

.EQU WS28XX_SPI_BIT_TICKS = 8
.EQU WS28XX_SPI_BYTE_TICKS = 8 * 8
.EQU WS28XX_BIT_TICKS = WS28XX_SPI_BIT_TICKS * WS28XX_ENC_BITS
.EQU WS28XX_CT_TICKS = WS28XX_SPI_BIT_TICKS * WS28XX_ENC_CT_BITS
.EQU WS28XX_FRAME_TICKS = WS28XX_CT_TICKS * (WS28XX_LEDCNT + WS28XX_RST_NCT)    

;informational: the frequencies ----------------------------------------------------------

.EQU WS28XX_SPICLOCK = FCLOCK / WS28XX_SPI_BIT_TICKS
.EQU WS28XX_INTRATE = FCLOCK / WS28XX_SPI_BYTE_TICKS
.EQU WS28XX_BITCLOCK = FCLOCK / WS28XX_BIT_TICKS
.EQU WS28XX_CTCLOCK = FCLOCK / WS28XX_CT_TICKS
.EQU WS28XX_FRAMERATE = FCLOCK / WS28XX_FRAME_TICKS

;variable data ---------------------------------------------------------------------------

.DSEG

ALIGN 32
ws28xx_data:
.BYTE WS28XX_ENC_CT_BYTES		;first buffer
ALIGN 16
.BYTE WS28XX_ENC_CT_BYTES		;second buffer

;include specific stuff ------------------------------------------------------------------

.INCLUDE "ws28xx_compat.inc"
.INCLUDE "ws28xx_tables.asm"
.INCLUDE "ws28xx_userdata.asm"

;Init code -------------------------------------------------------------------------------

.CSEG

ws28xx_init:
	sbi DDRD,PORTD1						;force output to low when SPI-TX is disabled

	clr NULL							;set up NULL register

	rcall ws28xx_userinit				;call user's init code

	ldi PTRL,Low(ws28xx_data)			;clear the buffers to start with a  
	ldi PTRH,High(ws28xx_data)			;refresh phase
	ldi R16,WS28XX_ENC_CT_BYTES
	_ws28xx_bufclr:
	st PTR,NULL
	adiw PTR,16
	st PTR+,NULL
	sbiw PTR,16	
	dec R16
	brne _ws28xx_bufclr
	sbiw PTR,1							;point to topmost byte of first buffer

	ldi R16,3							;setup USART for MSB first SPI mode
	UOUT WS28XX_UBRRL,R16				;with 2.5MBit/s -> WS28xx bitrate=833kHz
	ldi R16,(1<<WS28XX_UMSEL1)|(1<<WS28XX_UMSEL0)
	UOUT WS28XX_UCSRC,R16
	ldi R16,(1<<WS28XX_UDRIE)|(1<<WS28XX_TXEN)
	UOUT WS28XX_UCSRB,R16

	ldi LEDCNTL,Low(-WS28XX_RST_NCT)	;we start with the output of the reset phase
	ldi LEDCNTH,High(-WS28XX_RST_NCT)	;after (never sent) frame no. -1

	ret

;SPI USART UDRE ISR----------------------------------------------------------------------

SETVECTOR WS28XX_UDREAddr				; [6] [6]		constant interrupt latency + rjmp
	in SREGBAK,SREG						;  1   1		save flags

	ld TMP,PTR							;  2   2		retrieve current byte
	UOUT WS28XX_UDR,TMP					;  1   1		out to shift register
	subi PTRL,1							;  1   1		next byte
	brhs ws28xx_isr_bufswap				;  1   2		handle buffer swapping

	out SREG,SREGBAK					;  1			restore flags 
	reti								;  4			leave isr
										; --  ..
										; 17  13

;this is the end of the "normal" ISR, running in 8 of 9 cases ---------------------------
									
ws28xx_isr_bufswap:						;[13][13][13]
	andi PTRL,$18						;  1   1   1	fiddle pointer to point to
	ori PTRL,Low(ws28xx_data)			;  1   1   1	end of alternate buffer	  

	push YL								;  2   2   2	extra pointer register
	push YH								;  2   2   2    pair required
	push SREGBAK						;  2   2   2	release reserved register  

	ldi YL,$10							;  1   1   1	let the lower part point 
	eor YL,PTRL							;  1   1   1	to alternate-alternate buffer

	adiw LEDCNT,1						;  2   2   2	one LED (or refresh unit) progress
	brmi ws28xx_isr_inrefresh			;  1   1   2	already in refresh (only possible when WS28XX_RST_NCT>1)		 

	ldi YH,High(WS28XX_LEDCNT)			;  1   1		else compare count to max
	cpi LEDCNTL,Low(WS28XX_LEDCNT)		;  1   1
	cpc LEDCNTH,YH						;  1   1
	brcs ws28xx_isr_normalop			;  2   1
										; ..  ..  ..
										; 31  30  27

;this are the refresh cases for bufswap -------------------------------------------------

										;    [30][27]
	ldi LEDCNTL,Low(-WS28XX_RST_NCT)	;      1   			enter refresh phase
	ldi LEDCNTH,High(-WS28XX_RST_NCT)	;      1

ws28xx_isr_inrefresh:
	sei									;      1   1		leave exclusive execution, but
	mov YH,PTRH							;      1   1		this instruction is still protected
										;     ..  ..
										;     34  29 ->		end of exclusive code execution

										;    [34][29]
	st Y,NULL							;      2   2		the most important job in refresh phase 
	st -Y,NULL							;      2   2		clear the next buffer
	st -Y,NULL							;      2   2
	st -Y,NULL							;      2   2
	st -Y,NULL							;      2   2
	st -Y,NULL							;      2   2
	st -Y,NULL							;      2   2
	st -Y,NULL							;      2   2
	st -Y,NULL							;      2   2

	pop YL								;      2   2		get initial flags back
	sbr YL,1<<SREG_I 					;      1   1		we aren't exclusive at this point
	out SREG,YL							;      1   1		but have to restore remaining flags anyway

	pop YH								;      2   2		restore extra pointer 	
	pop YL								;      2   2		register pair
	ret									;      4   4		return from detached ISR 
										;     --  --
										;     64  59

;the normal case for bufswap: let the usercode fill the next buffer ----------------------

ws28xx_isr_normalop:					;[31]
	sei									;  1				leave exclusive execution, but
	push R0								;  2				this instruction is still protected			
										; ..
										; 34 ->				end of exclusive code execution

										;[34]
	push R1								;  2				some more registers required here
	push ZL								;  2
	push ZH								;  2
										; ..
										; 40

.IFDEF RAMPZ
	UIN ZH,RAMPZ
	push ZH
	.WARNING "3 extra ticks (more than 64k flash)"
.ENDIF

.INCLUDE "ws28xx_usercode.asm"			;???				let the user deliver LED data (R:R0,G:R1,B:YH)

.IFDEF RAMPZ
	ldi ZH,Byte3(ws28xx_enctables)
	UOUT RAMPZ,ZH
	.WARNING "2 extra ticks (more than 64k flash)"
.ENDIF

.IF WS28XX_RGBFMT == WS28XX_RGBFMT_RGB

	mov ZL,R0							;  1				encode RED value

	mov R0,YH							;  1				save BLUE value for later use
	ldi YH,High(ws28xx_data)			;  1				point to output buffer

	ldi ZH,High(ws28xx_enctables)+0		;  1
	ULPM ZH,Z							;  3
	st Y,ZH								;  2
	ldi ZH,High(ws28xx_enctables)+1		;  1
	ULPM ZH,Z							;  3
	st -Y,ZH							;  2
	ldi ZH,High(ws28xx_enctables)+2		;  1
	ULPM ZH,Z							;  3
	st -Y,ZH							;  2

	mov ZL,R1							;  1				encode GREEN value

	ldi ZH,High(ws28xx_enctables)+0		;  1
	ULPM ZH,Z							;  3
	st -Y,ZH							;  2
	ldi ZH,High(ws28xx_enctables)+1		;  1
	ULPM ZH,Z							;  3
	st -Y,ZH							;  2
	ldi ZH,High(ws28xx_enctables)+2		;  1
	ULPM ZH,Z							;  3
	st -Y,ZH							;  2

	mov ZL,R0							;  1				encode BLUE value

	ldi ZH,High(ws28xx_enctables)+0		;  1
	ULPM ZH,Z							;  3
	st -Y,ZH							;  2
	ldi ZH,High(ws28xx_enctables)+1		;  1
	ULPM ZH,Z							;  3
	st -Y,ZH							;  2
	ldi ZH,High(ws28xx_enctables)+2		;  1
	ULPM ZH,Z							;  3
	st -Y,ZH							;  2
										; ..
										; 59

.ELIF WS28XX_RGBFMT == WS28XX_RGBFMT_GRB

	mov ZL,R1							;  1				encode GREEN value

	mov R1,YH							;  1				save BLUE value for later use
	ldi YH,High(ws28xx_data)			;  1				point to output buffer

	ldi ZH,High(ws28xx_enctables)+0		;  1
	ULPM ZH,Z							;  3
	st Y,ZH								;  2
	ldi ZH,High(ws28xx_enctables)+1		;  1
	ULPM ZH,Z							;  3
	st -Y,ZH							;  2
	ldi ZH,High(ws28xx_enctables)+2		;  1
	ULPM ZH,Z							;  3
	st -Y,ZH							;  2

	mov ZL,R0							;  1				encode RED value

	ldi ZH,High(ws28xx_enctables)+0		;  1
	ULPM ZH,Z							;  3
	st -Y,ZH							;  2
	ldi ZH,High(ws28xx_enctables)+1		;  1
	ULPM ZH,Z							;  3
	st -Y,ZH							;  2
	ldi ZH,High(ws28xx_enctables)+2		;  1
	ULPM ZH,Z							;  3
	st -Y,ZH							;  2

	mov ZL,R1							;  1				encode BLUE value

	ldi ZH,High(ws28xx_enctables)+0		;  1
	ULPM ZH,Z							;  3
	st -Y,ZH							;  2
	ldi ZH,High(ws28xx_enctables)+1		;  1
	ULPM ZH,Z							;  3
	st -Y,ZH							;  2
	ldi ZH,High(ws28xx_enctables)+2		;  1
	ULPM ZH,Z							;  3
	st -Y,ZH							;  2
										; ..
										; 59
.ELSE

	.ERROR "RGB format not supported"

.ENDIF

.IFDEF RAMPZ
	pop ZH
	UOUT RAMPZ,ZH
	.WARNING "3 extra ticks (more than 64k flash)"
.ENDIF
										;[99]
	pop ZH								;  2				restore registers
	pop ZL								;  2
	pop R1								;  2
	pop R0								;  2

	pop YL								;  2				get initial flags back
	sbr YL,1<<SREG_I 					;  1		    	we aren't exclusive at this point
	out SREG,YL							;  1				but restore remaining flags

	pop YH								;  2				restore extra pointer 	
	pop YL								;  2				register pair
	ret									;  4				return from detached ISR 
										;---
										;119


;########################################################################################

;that means:
;
;(9 * 64) - (8 * 17) - 119 = 321
;
;321 ticks per 576 ticks long color triple cycle remains for user's ISR code, concurrencing
;interrupts and code in main. This is 55.7% of total computing time, virtually 11.1MHz
;system clock.

;The longest blocking time (effective as variable latency to all other ISRs in system) is
;34 ticks, equal to 1.7s. 

.ENDIF
